#define _XOPEN_SOURCE
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <utime.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

static char c_flag;
static char m_flag;
static char a_flag;

static int setdate(const char *path, const time_t date, char flag) {
	struct utimbuf new_time;

	if (flag || a_flag)
		new_time.actime = date;

	if (flag || m_flag)
		new_time.modtime = date;

	if (utime(path, &new_time) < 0) {
		fprintf(stderr, "touch: %s: %s\n", path, strerror(errno));
		return 1;
	}

	return 0;
}

static int touch(const char *path, const time_t date) {

	if (!c_flag) {
		int fd = open(path, O_CREAT | O_RDONLY, 0666);
		if (fd < 0) {
			fprintf(stderr, "touch: %s\n", strerror(errno));
			return 1;
		}

		close(fd);
	}

	/* Update the date to now */
	if (setdate(path, time(NULL), 1))
		return 1;

	/* Seting custom values */
	char flag = 0;
	if (m_flag == 0 && a_flag == 0)
		flag = 1;

	if (date != -1 && setdate(path, date, flag))
		return 1;

	return 0;
}

static time_t parse_date(char *str) {
	char *fmt = NULL;
	switch (strlen(str)) {
		case 8:
			fmt = "%m%d%H%M";
			break;

		case 10:
			fmt = "%y%m%d%H%M";
			break;

		case 11:
			fmt = "%m%d%H%M.%S";
			break;

		case 12:
			fmt = "%Y%m%d%H%M";
			break;

		case 13:
			fmt = "%y%m%d%H%M.%S";
			break;

		case 15:
			fmt = "%Y%m%d%H%M.%S";
			break;

		case 19:
			fmt = "%Y-%m-%dT%H:%M:%S";
			break;

		default:
			fprintf(stderr, "touch: invalid date format\n");
			exit(1);
	}

	time_t now = time(NULL);
	struct tm tm = *localtime(&now);

	if (!strptime(str, fmt, &tm)) {
		fprintf(stderr, "touch: %s\n", strerror(errno));
		exit(1);
	}

	return mktime(&tm);
}

int main(int argc, char **argv) {
	time_t date = -1;

	int opt;
	while ((opt = getopt(argc, argv, "cmad:t:")) != -1) {
		switch (opt) {
			case 'm':
				m_flag = 1;
				break;

			case 'a':
				a_flag = 1;
				break;

			case 'c':
				c_flag = 1;
				break;

			case 't':
			case 'd':
				date = parse_date(optarg);
				break;

			default:
				puts("touch [cmad] [file1 file2...]\n\t-c Don't create files\n\t-a Change only atime\n\t-m Change only mtime\n\t-d Date/time to use");
				return 0;
		}
	}

	argv += optind;
	argc -= optind;

	if (argc == 0) {
		printf("touch: missing operand\n");
		return 1;
	}

	for (int i = 0; i < argc; i++)
		if (touch(argv[i], date))
			return 1;

	return 0;
}
